# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
import numpy as np


def getXYval(data, X, Y, Z):
    assert Y.shape == X.shape, f"X shape {X.shape} does not match: Y shape {Y.shape}"
    dt = np.zeros(Y.shape)
    for i in range(Y.shape[0]):
        for j in range(Y.shape[1]):
            dt[i, j] = data[44 * 48 * abs(Z) + 44 * Y[i, j] + X[i, j]]
    return dt


def getYZval(data, X, Y, Z):
    assert Y.shape == Z.shape, f"Z shape {Z.shape} does not match: Y shape {Y.shape}"
    dt = np.zeros(Y.shape)
    for i in range(Y.shape[0]):
        for j in range(Y.shape[1]):
            dt[i, j] = data[44 * 48 * abs(Z[i, j]) + 44 * Y[i, j] + X]
    return dt


def getXZval(data, X, Y, Z):
    assert X.shape == Z.shape, f"Z shape {Z.shape} does not match: X shape {X.shape}"
    dt = np.zeros(X.shape)
    for i in range(X.shape[0]):
        for j in range(X.shape[1]):
            dt[i, j] = data[44 * 48 * abs(Z[i, j]) + 44 * Y + X[i, j]]
    return dt


def boxsurface(Nx, Ny, Nz, data, view_type, title, label, filename):
    X, Y, Z = np.meshgrid(np.arange(Nx), np.arange(Ny), -np.arange(Nz), indexing='ij')
    kw = {
        'vmin': data.min(),
        'vmax': data.max(),
        'levels': np.linspace(data.min(), data.max(), 10),
    }

    # Create a figure with 3D ax
    fig = plt.figure(figsize=(9, 9))
    ax = fig.add_subplot(111, projection='3d')
    # Set labels and zticks
    ax.set(
        xlabel='X Voxel',
        ylabel='Y Voxel',
        zlabel='Z Voxel',
    )
    # Set limits of the plot from coord limits
    xmin, xmax = X.min(), X.max()
    ymin, ymax = Y.min(), Y.max()
    zmin, zmax = Z.min(), Z.max()
    ax.set(xlim=[xmin, xmax], ylim=[ymin, ymax], zlim=[zmin, zmax])

    if view_type == 1:
        # Plot contour surfaces
        z_x = X[:, :, 0]
        z_y = Y[:, :, 0]
        z_offset = 0
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, -1, :]
        y_z = Z[:, -1, :]
        y_offset = Y.max()
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[-1, :, :]
        x_z = Z[-1, :, :]
        x_offset = X.max()
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=X.max(), **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmax, xmin], [ymax, ymax], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymax], [zmin, zmax], **edges_kw)
        # Set zoom and angle view
        ax.view_init(40, 30, 0)
    elif view_type == 2:
        # Plot contour surfaces
        z_x = X[:, :, 0]
        z_y = Y[:, :, 0]
        z_offset = 0
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, 0, :]
        y_z = Z[:, 0, :]
        y_offset = 0
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[-1, :, :]
        x_z = Z[-1, :, :]
        x_offset = X.max()
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=x_offset, **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmin, xmax], [ymin, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymin, ymin], [zmin, zmax], **edges_kw)
        # Set zoom and angle view
        ax.view_init(40, -30, 0)
    elif view_type == 3:
        # Plot contour surfaces
        z_x = X[:, :, -1]
        z_y = Y[:, :, -1]
        z_offset = Z.min()
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, -1, :]
        y_z = Z[:, -1, :]
        y_offset = Y.max()
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[-1, :, :]
        x_z = Z[-1, :, :]
        x_offset = X.max()
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=x_offset, **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmax, xmax], [ymax, ymax], [zmin, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymin], [zmin, zmin], **edges_kw)
        ax.plot([xmax, xmin], [ymax, ymax], [zmin, zmin], **edges_kw)
        # Set zoom and angle view
        ax.view_init(-40, 30, 0)
    elif view_type == 4:
        # Plot contour surfaces
        z_x = X[:, :, -1]
        z_y = Y[:, :, -1]
        z_offset = Z.min()
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, 0, :]
        y_z = Z[:, 0, :]
        y_offset = 0
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[-1, :, :]
        x_z = Z[-1, :, :]
        x_offset = X.max()
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=x_offset, **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmax, xmax], [ymax, ymin], [zmin, zmin], **edges_kw)
        ax.plot([xmin, xmax], [ymin, ymin], [zmin, zmin], **edges_kw)
        ax.plot([xmax, xmax], [ymin, ymin], [zmax, zmin], **edges_kw)
        # Set zoom and angle view
        ax.view_init(-40, -30, 0)
    elif view_type == 5:
        # Plot contour surfaces
        z_x = X[:, :, 0]
        z_y = Y[:, :, 0]
        z_offset = 0
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, -1, :]
        y_z = Z[:, -1, :]
        y_offset = Y.max()
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[0, :, :]
        x_z = Z[0, :, :]
        x_offset = 0
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=x_offset, **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmin, xmax], [ymax, ymax], [zmax, zmax], **edges_kw)
        ax.plot([xmin, xmin], [ymax, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmin, xmin], [ymax, ymax], [zmax, zmin], **edges_kw)
        # Set zoom and angle view
        ax.view_init(40, 120, 0)
    elif view_type == 6:
        # Plot contour surfaces
        z_x = X[:, :, 0]
        z_y = Y[:, :, 0]
        z_offset = 0
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, 0, :]
        y_z = Z[:, 0, :]
        y_offset = 0
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[0, :, :]
        x_z = Z[0, :, :]
        x_offset = 0
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=x_offset, **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmin, xmin], [ymax, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmin], [ymin, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmin, xmin], [ymin, ymin], [zmin, zmax], **edges_kw)
        # Set zoom and angle view
        ax.view_init(40, -120, 0)
    elif view_type == 7:
        # Plot contour surfaces
        z_x = X[:, :, -1]
        z_y = Y[:, :, -1]
        z_offset = Z.min()
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, -1, :]
        y_z = Z[:, -1, :]
        y_offset = Y.max()
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[0, :, :]
        x_z = Z[0, :, :]
        x_offset = 0
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=x_offset, **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmin, xmin], [ymax, ymax], [zmax, zmin], **edges_kw)
        ax.plot([xmax, xmin], [ymax, ymax], [zmin, zmin], **edges_kw)
        ax.plot([xmin, xmin], [ymin, ymax], [zmin, zmin], **edges_kw)
        # Set zoom and angle view
        ax.view_init(-40, 120, 0)
    elif view_type == 8:
        # Plot contour surfaces
        z_x = X[:, :, -1]
        z_y = Y[:, :, -1]
        z_offset = Z.min()
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, 0, :]
        y_z = Z[:, 0, :]
        y_offset = 0
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[0, :, :]
        x_z = Z[0, :, :]
        x_offset = 0
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=x_offset, **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmin, xmin], [ymax, ymin], [zmin, zmin], **edges_kw)
        ax.plot([xmin, xmin], [ymin, ymin], [zmax, zmin], **edges_kw)
        ax.plot([xmax, xmin], [ymin, ymin], [zmin, zmin], **edges_kw)
        # Set zoom and angle view
        ax.view_init(-40, -120, 0)

    ax.set_box_aspect(None, zoom=1)

    # Colorbar
    fig.colorbar(A, ax=ax, fraction=0.02, pad=0.1, label=label)
    ax.set_title(title)
    plt.savefig(filename)
    # Show Figure
    #plt.show()
    plt.close('all')


def boxCrossSection(Nx, Ny, Nz, data, view_type, offset, title, label, filename):
    X, Y, Z = np.meshgrid(np.arange(Nx), np.arange(Ny), -np.arange(Nz), indexing='ij')
    kw = {
        'vmin': 0,
        'vmax': data.max(),
        'levels': np.linspace(data.min(), data.max(), 5),
    }

    # Create a figure with 3D ax
    fig = plt.figure(figsize=(9, 9))
    ax = fig.add_subplot(111, projection='3d')
    # Set labels and zticks
    ax.set(
        xlabel='X Voxel',
        ylabel='Y Voxel',
        zlabel='Z Voxel',
    )
    # Set limits of the plot from coord limits
    xmin, xmax = X.min(), X.max()
    ymin, ymax = Y.min(), Y.max()
    zmin, zmax = Z.min(), Z.max()
    ax.set(xlim=[xmin, xmax], ylim=[ymin, ymax], zlim=[zmin, zmax])
    if view_type == 1:
        # Plot contour surfaces
        z_x = X[0:offset, :, 0]
        z_y = Y[0:offset, :, 0]
        z_offset = 0
        z_d = getXYval(data, z_x, z_y, z_offset)
        z_d[z_d == 0] = np.nan
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[0:offset, -1, :]
        y_z = Z[0:offset, -1, :]
        y_offset = Y.max()
        y_d = getXZval(data, y_x, y_offset, y_z)
        y_d[y_d == 0] = np.nan
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[offset, :, :]
        x_z = Z[offset, :, :]
        x_offset = offset - 1
        x_d = getYZval(data, x_offset, x_y, x_z)
        x_d[x_d == 0] = np.nan
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=x_offset, **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmax, xmin], [ymax, ymax], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymax], [zmin, zmax], **edges_kw)
        ax.plot([offset - 1, offset - 1], [ymax, ymax], [zmin, zmax], **edges_kw)
        ax.plot([offset - 1, offset - 1], [ymin, ymax], [zmax, zmax], **edges_kw)
        # Set zoom and angle view
        ax.view_init(40, 30, 0)
    elif view_type == 2:
        z_x = X[:, 0:offset, 0]
        z_y = Y[:, 0:offset, 0]
        z_offset = 0
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, offset, :]
        y_z = Z[:, offset, :]
        y_offset = offset - 1
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[-1, 0:offset, :]
        x_z = Z[-1, 0:offset, :]
        x_offset = X.max()
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=X.max(), **kw)
        # Plot edges
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmax, xmin], [ymax, ymax], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymax], [zmin, zmax], **edges_kw)
        ax.plot([xmax, xmax], [offset - 1, offset - 1], [zmax, zmin], **edges_kw)
        ax.plot([xmin, xmax], [offset - 1, offset - 1], [zmax, zmax], **edges_kw)
        # Set zoom and angle view
        ax.view_init(40, 30, 0)
    elif view_type == 3:
        # Plot contour surfaces
        z_x = X[:, :, offset]
        z_y = Y[:, :, offset]
        z_offset = -offset
        z_d = getXYval(data, z_x, z_y, z_offset)
        A = ax.contourf(z_x, z_y, z_d, zdir='z', offset=z_offset, **kw)

        y_x = X[:, -1, offset:17]
        y_z = Z[:, -1, offset:17]
        y_offset = Y.max()
        y_d = getXZval(data, y_x, y_offset, y_z)
        B = ax.contourf(y_x, y_d, y_z, zdir='y', offset=y_offset, **kw)

        x_y = Y[-1, :, offset:17]
        x_z = Z[-1, :, offset:17]
        x_offset = X.max()
        x_d = getYZval(data, x_offset, x_y, x_z)
        C = ax.contourf(x_d, x_y, x_z, zdir='x', offset=X.max(), **kw)
        edges_kw = dict(color='0.7', linewidth=1, zorder=1e3)
        ax.plot([xmax, xmin], [ymax, ymax], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymin], [zmax, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymax, ymax], [zmin, zmax], **edges_kw)
        ax.plot([xmax, xmax], [ymin, ymax], [zmax - offset, zmax - offset], **edges_kw)
        ax.plot([xmin, xmax], [ymax, ymax], [zmax - offset, zmax - offset], **edges_kw)
        # ax.plot([xmax, xmax], [ymax, ymax], [zmin, zmax], **edges_kw)
        # Set zoom and angle view
        ax.view_init(40, 30, 0)

    ax.set_box_aspect(None, zoom=1)

    # Colorbar
    fig.colorbar(A, ax=ax, fraction=0.02, pad=0.1, label=label)
    ax.set_title(title)
    plt.savefig(filename)
    # Show Figure
    #plt.show()
    plt.close('all')
