# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import os.path
import argparse
import constants as ct
import importandsave as ims
import numpy as np
import pandas as pd


def calculate3dVoxelIndices(xfile, yfile, zfile, resultfile, xsize=44, ysize=48, zsize=17):
    x = pd.read_csv(xfile)
    y = pd.read_csv(yfile)
    z = pd.read_csv(zfile)

    d_dx = np.array(x["DX"])
    d_dy = np.array(y["DY"])
    d_thick = np.array(z["Thickness"])
    zz, yy, xx = getIndices(d_dx, d_dy, d_thick, xsize, ysize, zsize)
    df = pd.DataFrame({"z": zz.flatten(), "y": yy.flatten(), "x": xx.flatten()})
    df.to_csv(resultfile)


def calculate3dVoxelMidPoints(xfile, yfile, zfile, resultfile, xsize=44, ysize=48, zsize=17):
    x = pd.read_csv(xfile)
    y = pd.read_csv(yfile)
    z = pd.read_csv(zfile)

    d_dx = np.array(x["DX"])
    d_dy = np.array(y["DY"])
    d_thick = np.array(z["Thickness"])
    xx, yy, zz = getMidPoints(d_dx, d_dy, d_thick, xsize, ysize, zsize)
    df = pd.DataFrame({"z": zz.flatten(), "y": yy.flatten(), "x": xx.flatten()})
    df.to_csv(resultfile)


def writeMidpoints(xfile, yfile, zfile, source, destination):
    x = pd.read_csv(xfile)
    y = pd.read_csv(yfile)
    z = pd.read_csv(zfile)

    d_dx = np.array(x["DX"])
    d_dy = np.array(y["DY"])
    d_thick = np.array(z["Thickness"])
    xx, yy, zz = getMidPoints(d_dx, d_dy, d_thick)

    for name in ct.namelist:
        filename = name + ".csv"
        fullpath = os.path.join(source, filename)
        df = pd.read_csv(fullpath)
        ind = ims.getTextIndex(ct.namelist, name)
        ims.setData(ind, ct.datalist, np.array(df[name]))

    for name in ct.shortnamelist:
        ind = ims.getTextIndex(ct.namelist, name)
        data = ims.getData(ind, ct.datalist)
        sfile = ct.filenamelist[ind]
        filename = sfile + ".csv"
        fullpath = os.path.join(destination, filename)
        dict_val = {"X": xx.flatten(), "Y": yy.flatten(), "Z": -zz.flatten(), name: data}
        df_val = pd.DataFrame(dict_val)
        df_val.to_csv(fullpath)


def getIndices(d_dx, d_dy, d_thick, xsize=44, ysize=48, zsize=17):
    dx = np.array(d_dx)
    dx = dx.reshape((ysize, xsize))
    print(dx[0, :])
    print(np.sum(dx[0, :]))

    dy = np.array(d_dy)
    dy = dy.reshape((ysize, xsize))
    print(dy[:, 0])
    print(np.sum(dy[:, 0]))

    dz = np.array(d_thick)
    dz = dz.reshape((zsize, ysize, xsize))
    print(dz[:, 0, 0])
    print(np.sum(dz[:, 0, 0]))

    x = dx[0, :]
    y = dy[:, 0]
    z = dz[:, 0, 0]

    zz, yy, xx = np.indices((zsize + 1, ysize + 1, xsize + 1))

    total = 0
    for i in range(xsize):
        if i != 0:
            xx[:, :, i] = xx[:, :, i] * (total / i)
        total = total + x[i]
    xx[:, :, xsize] = xx[:, :, xsize] * (total / xsize)

    total = 0
    for i in range(ysize):
        if i != 0:
            yy[:, i, :] = yy[:, i, :] * (total / i)
        total = total + y[i]
    yy[:, ysize, :] = yy[:, ysize, :] * (total / ysize)

    total = 0
    for i in range(zsize):
        s = zz[i, :, :]
        if i != 0:
            zz[i, :, :] = zz[i, :, :] * (total / i)
        s = zz[i, :, :]
        total = total + z[i]
    zz[zsize, :, :] = zz[zsize, :, :] * (total / 17)

    return zz, yy, xx


def getMidPoints(d_dx, d_dy, d_thick, xsize=44, ysize=48, zsize=17):
    dx = np.array(d_dx)
    dx = dx.reshape((ysize, xsize))
    print(dx[0, :])
    print(np.sum(dx[0, :]))

    dy = np.array(d_dy)
    dy = dy.reshape((ysize, xsize))
    print(dy[:, 0])
    print(np.sum(dy[:, 0]))

    dz = np.array(d_thick)
    dz = dz.reshape((zsize, ysize, xsize))
    print(dz[:, 0, 0])
    print(np.sum(dz[:, 0, 0]))

    x = dx[0, :]
    y = dy[:, 0]
    z = dz[:, 0, 0]

    xx = np.zeros((zsize, ysize, xsize))
    prev = 0
    for i in range(xsize):
        next = prev + x[i]
        xx[:, :, i] = (prev + next) / 2
        prev = prev + x[i]

    yy = np.zeros((zsize, ysize, xsize))
    prev = 0
    for i in range(ysize):
        next = prev + y[i]
        yy[:, i, :] = (prev + next) / 2
        prev = prev + y[i]

    zz = np.zeros((zsize, ysize, xsize))
    prev = 0
    for i in range(zsize):
        next = prev + z[i]
        zz[i, :, :] = (prev + next) / 2
        prev = prev + z[i]

    return xx, yy, zz


def compareWellData(leapwellcsv, d3dwellcsv, bothwells, leaponly, d3donly):
    wf1 = pd.read_csv(leapwellcsv)
    wf2 = pd.read_csv(d3dwellcsv)

    wn1 = wf1["Well_Name"]
    wn2 = wf2["Name"]

    both = []
    for w1 in wn1:
        if w1 in wn2.values:
            both.append(w1)
    pd.DataFrame(both).to_csv(bothwells)

    first = []
    for w1 in wn1:
        if not w1 in wn2.values:
            first.append(w1)
    pd.DataFrame(first).to_csv(leaponly)

    second = []
    for w2 in wn2:
        if not w2 in wn1.values:
            second.append(w2)
    pd.DataFrame(second).to_csv(d3donly)


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument("-i", "--indices", required=False,
                    help="flag to create voxel indices and export to output csv file", action='store_true',
                    default=False)
    ap.add_argument("-m", "--midpoints", required=False,
                    help="flag to create voxel midpoints and export to output csv file", action='store_true',
                    default=False)
    ap.add_argument("-a", "--allmidpoints", required=False,
                    help="flag to create voxel midpoints together with all values in source folder and export to output csv file",
                    action='store_true', default=False)
    ap.add_argument("-x", "--xfile", required=False,
                    help="csv file containing x dimension of each voxel with column name DX", default=None)
    ap.add_argument("-y", "--yfile", required=False,
                    help="csv file containing y dimension of each voxel with column name DY", default=None)
    ap.add_argument("-z", "--zfile", required=False,
                    help="csv file containing z dimension of each voxel with column name Thickness", default=None)
    ap.add_argument("-o", "--output", required=False, help="Destination csv file", default=None)
    ap.add_argument("-s", "--source", required=False, help="Source csv file", default=None)
    ap.add_argument("-d", "--destination", required=False, help="Destination folder to write csv files", default=None)
    ap.add_argument("-w", "--well", required=False,
                    help="flag to compare well data in two csv files and write the result in 3 csv files (intersection, firstonly, and secondonly)", action='store_true',
                    default=False)
    ap.add_argument("-w1", "--well1", required=False,
                    help="first well input csv file", default=None)
    ap.add_argument("-w2", "--well2", required=False,
                    help="second well input csv file", default=None)
    ap.add_argument("-w3", "--well3", required=False,
                    help="first well output csv file (intersection)", default=None)
    ap.add_argument("-w4", "--well4", required=False,
                    help="second well output csv file (firstonly)", default=None)
    ap.add_argument("-w5", "--well5", required=False,
                    help="third well output csv file (secondonly)", default=None)
    args = vars(ap.parse_args())
    ind = args["indices"]
    mid = args["midpoints"]
    all = args["allmidpoints"]
    x = args["xfile"]
    y = args["yfile"]
    z = args["zfile"]
    src = args["source"]
    dst = args["destination"]
    out = args["output"]
    wll = args["well"]
    wll1 = args["well1"]
    wll2 = args["well2"]
    wll3 = args["well3"]
    wll4 = args["well4"]
    wll5 = args["well5"]
    if ind and (not mid) and (not all) and (out is not None) and x is not None and y is not None and z is not None:
        calculate3dVoxelIndices(x, y, z, out)
    elif (not ind) and mid and (not all) and (out is not None) and x is not None and y is not None and z is not None:
        calculate3dVoxelMidPoints(x, y, z, out)
    elif (not ind) and (not mid) and all and (src is not None) and (dst is not None) and x is not None and y is not None and z is not None:
        writeMidpoints(x, y, z, src, dst)
    elif wll and wll1 is not None and wll2 is not None and wll3 is not None and wll4 is not None and wll5 is not None:
        compareWellData(wll1, wll2, wll3, wll4, wll5)
    else:
        print("wrong parameters")
