# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import os.path
import argparse
import constants as ct
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import importandsave as ims
import calculate as cl
from boxsurface import boxsurface, boxCrossSection


def draw3Dgraphs(source, indices, destination, xsize=44, ysize=48, zsize=17):
    indfile = pd.read_csv(indices)
    x = indfile["x"]
    y = indfile["y"]
    z = indfile["z"]
    xx = np.array(x).reshape(zsize + 1, ysize + 1, xsize + 1)
    yy = np.array(y).reshape(zsize + 1, ysize + 1, xsize + 1)
    zz = np.array(z).reshape(zsize + 1, ysize + 1, xsize + 1)

    for name in ct.namelist:
        filename = name + ".csv"
        fullpath = os.path.join(source, filename)
        df = pd.read_csv(fullpath)
        ind = ims.getTextIndex(ct.namelist, name)
        ims.setData(ind, ct.datalist, np.array(df[name]))

    for name in ct.shortnamelist:
        b_values = ims.getValues(name, ct.namelist, ct.datalist)
        ind = ims.getTextIndex(ct.namelist, name)
        data = ims.getData(ind, ct.datalist)
        data_3d = np.array(data).reshape((zsize, ysize, xsize))
        stitle = ct.titlelist[ind]
        sfile = ct.filenamelist[ind]
        for val in b_values:
            title = stitle + " = " + str(val)
            filename = sfile + "_equals_" + str(val)
            directory = os.path.join(destination, sfile, "equals")
            isExist = os.path.exists(directory)
            if not isExist:
                os.makedirs(directory)
            fullpath = os.path.join(directory, filename)
            draw3d(data_3d, val, title, "equals", fullpath, zz, yy, xx)
            print(name, " equals ", val)
            title = stitle + " > " + str(val)
            filename = sfile + "_greater_" + str(val)
            directory = os.path.join(destination, sfile, "greater")
            isExist = os.path.exists(directory)
            if not isExist:
                os.makedirs(directory)
            fullpath = os.path.join(destination, filename)
            draw3d(data_3d, val, title, "greater", fullpath, zz, yy, xx)
            print(name, " greater ", val)
            title = stitle + " < " + str(val)
            filename = sfile + "_less_" + str(val)
            directory = os.path.join(destination, sfile, "less")
            isExist = os.path.exists(directory)
            if not isExist:
                os.makedirs(directory)
            fullpath = os.path.join(destination, filename)
            draw3d(data_3d, val, title, "less", fullpath, zz, yy, xx)
            print(name, " less ", val)
            plt.close('all')


def draw3d(data, value, title, comp, filename, zz, yy, xx):
    list = [(40, 30), (-40, 30), (40, -30), (-40, -30)]
    count = 1
    for ls in list:
        plt.subplot(2, 2, count)
        zzz = -zz
        x, y, z = np.indices((17, 48, 44))

        # define a sphere about [0.5, 0.5, 0.5]
        valid = None
        if comp == "equals":
            valid = data == value
        elif comp == "greater":
            valid = data > value
        elif comp == "less":
            valid = data < value

        # combine the color components
        colors = np.zeros((44, 48, 17) + (3,))

        colors[..., 0] = x.T / 60
        colors[..., 1] = y.T / 60
        colors[..., 2] = z.T / 60

        edgecolor = np.clip(2 * colors - 0.5, 0, 1)

        valid = valid.T

        # and plot everything
        ax = plt.figure().add_subplot(projection='3d')
        ax.voxels(xx.T, yy.T, zzz.T, valid,
                  facecolors=colors,
                  edgecolors=edgecolor,  # brighter
                  linewidth=0.5)
        ax.set(xlabel='x', ylabel='y', zlabel='z')
        ax.set_title(title)
        ax.set_aspect('equal')
        ax.view_init(ls[0], ls[1], 0)
        plt.savefig(filename + "-" + str(count) + ".png")
        # plt.show()
        count = count + 1


def drawBoxSurfaces(source, destination, xsize=44, ysize=48, zsize=17):
    for name in ct.namelist:
        filename = name + ".csv"
        fullpath = os.path.join(source, filename)
        df = pd.read_csv(fullpath)
        ind = ims.getTextIndex(ct.namelist, name)
        ims.setData(ind, ct.datalist, np.array(df[name]))

    for name in ct.shortnamelist:
        ind = ims.getTextIndex(ct.namelist, name)
        data = ims.getData(ind, ct.datalist)
        stitle = ct.titlelist[ind]
        sfile = ct.filenamelist[ind]
        label = ct.labellist[ind]
        directory = os.path.join(destination, name)
        isExist = os.path.exists(directory)
        if not isExist:
            os.makedirs(directory)
        sfile = os.path.join(directory, filename)
        print(name)
        boxsurface(xsize, ysize, zsize, np.array(data), 1, stitle, label, sfile + "_1.png")
        boxsurface(xsize, ysize, zsize, np.array(data), 2, stitle, label, sfile + "_2.png")
        boxsurface(xsize, ysize, zsize, np.array(data), 3, stitle, label, sfile + "_3.png")
        boxsurface(xsize, ysize, zsize, np.array(data), 4, stitle, label, sfile + "_4.png")
        boxsurface(xsize, ysize, zsize, np.array(data), 5, stitle, label, sfile + "_5.png")
        boxsurface(xsize, ysize, zsize, np.array(data), 6, stitle, label, sfile + "_6.png")
        boxsurface(xsize, ysize, zsize, np.array(data), 7, stitle, label, sfile + "_7.png")
        boxsurface(xsize, ysize, zsize, np.array(data), 8, stitle, label, sfile + "_8.png")
        plt.close('all')


def drawBoxCrossSections(source, destination, xsize=44, ysize=48, zsize=17):
    for name in ct.namelist:
        filename = name + ".csv"
        fullpath = os.path.join(source, filename)
        df = pd.read_csv(fullpath)
        ind = ims.getTextIndex(ct.namelist, name)
        ims.setData(ind, ct.datalist, np.array(df[name]))

    for name in ct.shortnamelist:
        ind = ims.getTextIndex(ct.namelist, name)
        data = ims.getData(ind, ct.datalist)
        dt = np.array(data)
        stitle = ct.titlelist[ind]
        sfile = ct.filenamelist[ind]
        label = ct.labellist[ind]
        directory = os.path.join(destination, name)
        isExist = os.path.exists(directory)
        if not isExist:
            os.makedirs(directory)
        sfile = os.path.join(directory, sfile)
        print(name)
        for i in range(2, 44):
            boxCrossSection(xsize, ysize, zsize, dt, 1, i, stitle + " X Axis Cross Section X=" + str(i),
                            label, sfile + "_X_CS_" + str(i) + ".png")
        for i in range(2, 48):
            boxCrossSection(xsize, ysize, zsize, dt, 2, i, stitle + " Y Axis Cross Section Y=" + str(i),
                            label, sfile + "_Y_CS_" + str(i) + ".png")
        for i in range(0, 16):
            boxCrossSection(xsize, ysize, zsize, dt, 3, i, stitle + " Z Axis Cross Section Z=-" + str(i),
                            label, sfile + "_Z_CS_" + str(i) + ".png")


def drawWellsCrossSection(source, destination, xsize=44, ysize=48, zsize=17):
    wells = pd.read_csv(source)
    wells_data = np.array(wells["val"])
    isExist = os.path.exists(destination)
    if not isExist:
        os.makedirs(destination)
    file = os.path.join(destination, "Wells")
    for i in range(2, 44):
        boxCrossSection(xsize, ysize, zsize, wells_data, 1, i, "Wells X Axis Cross Section X=" + str(i),
                        "wells", file + "_X_CS_" + str(i))
    for i in range(2, 48):
        boxCrossSection(xsize, ysize, zsize, wells_data, 2, i, "Wells X Axis Cross Section X=" + str(i),
                        "wells", file + "_Y_CS_" + str(i))
    for i in range(0, 16):
        boxCrossSection(xsize, ysize, zsize, wells_data, 3, i, "Wells X Axis Cross Section X=" + str(i),
                        "wells", file + "_Z_CS_" + str(i))


""" 
def drawWells3DGraphs(voxelsource, wellsource, xsize=44, ysize=48, zsize=17):
    indfile = pd.read_csv("csv2015/3DVoxelIndices.csv")
    x = indfile["x"]
    y = indfile["y"]
    z = indfile["z"]
    xx = np.array(x).reshape(18, 49, 45)
    yy = np.array(y).reshape(18, 49, 45)
    zz = np.array(z).reshape(18, 49, 45)

    wells = pd.read_csv("csvWellLocations/wells_data.csv")
    wells_data = np.array(wells["val"])
    data_3d = np.array(wells_data).reshape((17, 48, 44))
    stitle = "Wells"
    sfile = "Wells"
    title = stitle + " = PRODUCTION"
    filename = sfile + "_equals_PRODUCTION"
    draw3d(data_3d, 1, title, "equals", filename, zz, yy, xx)
    title = stitle + " = INJECTION"
    filename = sfile + "_equals_INJECTION"
    draw3d(data_3d, 2, title, "equals", filename, zz, yy, xx)
    draw3d(data_3d, 0, title, "greater", sfile, zz, yy, xx)


def scatterPlotBothWells():
    x = pd.read_csv("csv2015/DX.csv")
    y = pd.read_csv("csv2015/DY.csv")
    z = pd.read_csv("csv2015/Thickness.csv")

    d_dx = np.array(x["DX"])
    d_dy = np.array(y["DY"])
    d_thick = np.array(z["Thickness"])
    xx, yy, zz = cl.getMidPoints(d_dx, d_dy, d_thick)

    both = pd.read_csv("csvWellLocations/both.csv")
    wells = pd.read_csv("csvWellLocations/wells.csv")

    X = []
    Y = []
    name = []
    for wn in both["0"]:
        found = wells[wells["Name"] == wn]
        if found.size > 0:
            str = found["Location"].iloc[0]
            val = str.replace("[", "").replace("]", "")
            ls = list(map(int, val.split(',')))
            arr = np.array(ls)
            val = (arr[0] - 1) % 2112
            X.append(xx.flatten()[val])
            Y.append(yy.flatten()[val])
            name.append(wn)
    fig, ax = plt.subplots()
    ax.scatter(X, Y, c="blue")

    for i, txt in enumerate(name):
        ax.annotate(txt, (X[i], Y[i]))

    plt.show()
"""


def draw3DgraphsFile(source, indices, destination, xsize=44, ysize=48, zsize=17):
    indfile = pd.read_csv("calculate/voxelindices.csv")
    data_3d = np.load("faultdensity/fault.npy")
    x = indfile["x"]
    y = indfile["y"]
    z = indfile["z"]
    xx = np.array(x).reshape(zsize + 1, ysize + 1, xsize + 1)
    yy = np.array(y).reshape(zsize + 1, ysize + 1, xsize + 1)
    zz = np.array(z).reshape(zsize + 1, ysize + 1, xsize + 1)

    draw3d(data_3d, 1, "title", "equals", "fault_equals", zz, yy, xx)
    plt.close('all')



if __name__ == '__main__':
    #draw3DgraphsFile(None, None,None)
    ap = argparse.ArgumentParser()
    ap.add_argument("-t", "--type", required=True,
                    help="drawing type. ",
                    default=None)
    ap.add_argument("-i", "--indices", required=False,
                    help="flag to create voxel indices and export to output csv file",
                    default=False)
    ap.add_argument("-s", "--source", required=False, help="Source csv file", default=None)
    ap.add_argument("-d", "--destination", required=False, help="Destination folder to write csv files", default=None)

    args = vars(ap.parse_args())
    typ = args["type"]
    ind = args["indices"]
    src = args["source"]
    dst = args["destination"]
    if typ == "3D":
        draw3Dgraphs(src, ind, dst)
    elif typ == "boxsurface":
        drawBoxSurfaces(src, dst)
    elif typ == "boxcrosssection":
        drawBoxCrossSections(src, dst)
    elif typ == "wellcrosssection":
        drawWellsCrossSection(src, dst)
    else:
        print("wrong parameters")
