# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import pandas as pd
import argparse
from osgeo import gdal
from tqdm import tqdm
import os


def getLithology(z, y, x, lyrs):
    for name, npyfile in lyrs:
        if npyfile[z][y][x] == 1:
            return name


def export(rasterfile, folder, layers, csvcolumnname, exportfile, zref=850, width=100):
    ds = gdal.Open(rasterfile)
    lyrs = []
    for layer in layers:
        name = layer.split(".")[0]
        npyfile = np.load(os.path.join(folder, layer))
        lss = [name, npyfile]
        lyrs.append(lss)
    gt = ds.GetGeoTransform()
    xref = gt[0]
    yref = gt[3]
    size = gt[1]
    center = size / 2
    X = []
    Y = []
    Z = []
    Pr = []
    for z in tqdm(range(lyrs[0][1].shape[0])):
        for y in range(lyrs[0][1].shape[1]):
            for x in range(lyrs[0][1].shape[2]):
                xval = xref + (x * size) + center
                yval = yref - (y * size) - center
                zval = zref - (z * size)
                X.append(xval)
                Y.append(yval)
                Z.append(zval)
                Pr.append(getLithology(z, y, x, lyrs))

    dict_val = {"X": np.array(X), "Y": np.array(Y), "Z": np.array(Z), csvcolumnname: np.array(Pr)}
    df_val = pd.DataFrame(dict_val)
    df_val.to_csv(exportfile, index=True, index_label="Id")
    print("Result CSV file exported to", exportfile)


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument("-r", "--raster", required=True, help="raster file", default=None)
    ap.add_argument("-f", "--folder", required=True, help="folder", default=None)
    ap.add_argument("-l", "--layers", required=True, help="input layers", nargs='+', default=[])
    ap.add_argument("-n", "--name", required=True, help="name of column", default=None)
    ap.add_argument("-e", "--export", required=True, help="export csv filename", default=None)
    args = vars(ap.parse_args())
    raster = args["raster"]
    folder = args["folder"]
    layers = args["layers"]
    name = args["name"]
    exp = args["export"]

    export(raster, folder, layers, name, exp)
