# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


from osgeo import gdal, ogr, osr
from tqdm import tqdm

import numpy as np
import pandas as pd
import math
import constants as ct
import argparse
import os


def create3DModelGrill(sourceshp, destshp, xsource, ysource, xmaxcount=44, epsg=4269, utm=11):
    driver = ogr.GetDriverByName("ESRI Shapefile")
    source = driver.Open(sourceshp, 0)
    layer = source.GetLayer()
    xmin, xmax, ymin, ymax = 9999999, 0, 9999999, 0
    ulx = uly = urx = ury = llx = lly = lrx = lry = 0
    for n in range(layer.GetFeatureCount()):
        feature = layer.GetFeature(n)
        geometry = feature.GetGeometryRef()
        ring = geometry.GetGeometryRef(0)
        print(geometry.GetGeometryName())
        points = ring.GetPointCount()
        for p in range(points):
            x, y, z = ring.GetPoint(p)
            print(x, y)
            if x > xmax:
                xmax = x
                urx = x
                ury = y
            if x < xmin:
                xmin = x
                llx = x
                lly = y
            if y > ymax:
                ymax = y
                uly = y
                ulx = x
            if y < ymin:
                ymin = y
                lrx = x
                lry = y
    print("ul", ulx, uly)
    print("ur", urx, ury)
    print("ll", llx, lly)
    print("lr", lrx, lry)

    xcoord = pd.read_csv(xsource)
    xcoord = np.array(xcoord["DX"])
    ycoord = pd.read_csv(ysource)
    ycoord = np.array(ycoord["DY"])
    xy = np.stack((xcoord, ycoord), axis=1)
    driver2 = ogr.GetDriverByName("ESRI Shapefile")
    ds = driver2.CreateDataSource(destshp)
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)
    srs.SetUTM(utm, True)
    # create one layer
    layer = ds.CreateLayer("polygon", srs, ogr.wkbPolygon)
    # Add an ID field
    idField = ogr.FieldDefn("id", ogr.OFTInteger)
    layer.CreateField(idField)
    featureDefn = layer.GetLayerDefn()

    xref = xref2 = ulx
    yref = yref2 = uly
    xm = getM(ulx, uly, urx, ury)
    ym = getM(ulx, uly, llx, lly)
    for vl in range(xy.shape[0]):
        x = xy[vl][0]
        y = xy[vl][1]
        xd = x * ct.FEET
        yd = y * ct.FEET
        xdist_x, ydist_x = getDist(xd, xm)
        xdist_y, ydist_y = getDist(yd, ym)
        ring = ogr.Geometry(ogr.wkbLinearRing)
        ring.AddPoint(xref, yref)
        ring.AddPoint(xref + xdist_x, yref - ydist_x)
        ring.AddPoint(xref + xdist_x - xdist_y, yref - ydist_x - ydist_y)
        ring.AddPoint(xref - xdist_y, yref - ydist_y)
        ring.AddPoint(xref, yref)
        # Create polygon
        poly = ogr.Geometry(ogr.wkbPolygon)
        poly.AddGeometry(ring)
        feature = ogr.Feature(featureDefn)
        feature.SetGeometry(poly)
        feature.SetField("id", vl + 1)
        layer.CreateFeature(feature)
        feature = None
        if (vl + 1) % xmaxcount != 0:
            xref = xref + xdist_x
            yref = yref - ydist_x
        else:
            xref = xref2 - xdist_y
            yref = yref2 - ydist_y
            xref2 = xref
            yref2 = yref
    ds = None


def getDist(d, m):
    m2 = m * m
    xval = abs(d / math.sqrt(1 + m2))
    yval = abs(d / math.sqrt(((m2 + 1) / m2)))
    return xval, yval


def getM(x1, y1, x2, y2):
    val = (y1 - y2) / (x1 - x2)
    return val


def write3DValuestoCSV(sourcerasterfile, thicknessfile, resultfile):
    ds = gdal.Open(sourcerasterfile)
    z_thick = pd.read_csv(thicknessfile)
    zarray = np.array(z_thick["Thickness"]).reshape(-1, 2112)
    zl = zarray[:, 0]
    gt = ds.GetGeoTransform()
    data = ds.GetRasterBand(1)
    dt = data.ReadAsArray()
    xref = gt[0]
    yref = gt[3]
    zref = ct.Z_REF
    size = gt[1]
    center = size / 2
    zlenght = 12300
    zlenght = zlenght * ct.FEET
    zsize = int(zlenght / size) + 1
    list = []

    for z in tqdm(range(zsize)):
        for y in range(dt.shape[0]):
            for x in range(dt.shape[1]):
                ls = []
                xval = xref + (x * size) + center
                yval = yref - (y * size) - center
                zval = zref - (z * size)
                vl = dt[y][x]
                if vl != 65535:
                    zp = getVoxelZ(z, zl, size)
                    vl = vl + (zp * 2112)
                ls.append(xval)
                ls.append(yval)
                ls.append(zval)
                ls.append(vl)
                list.append(ls)

    csv = pd.DataFrame(list, columns=["X", "Y", "Z", "VoxelID"])
    csv.to_csv(resultfile)


def getVoxelZ(step, zl, size):
    total = 0
    for i in range(len(zl)):
        total = total + zl[i]
        if (step * size) < (total * ct.FEET):
            return i


def createInput2DFiles(rasterfile, sourcefolder, destinationfolder):
    ds = gdal.Open(rasterfile)
    driver = gdal.GetDriverByName('GTiff')
    gt = ds.GetGeoTransform()
    crs = ds.GetProjection()
    cols = ds.RasterXSize
    rows = ds.RasterYSize
    for filename in os.listdir(sourcefolder):
        if ".npy" in filename:
            print("Processing Data: ", filename)
            file = os.path.join(sourcefolder, filename)
            array_data = np.load(file)
            for z in tqdm(range(array_data.shape[0])):
                data = array_data[z, :, :]
                min = np.nanmin(data)
                max = np.nanmax(data)
                print(file, z, min, max)
                name = filename.replace(".npy", "")
                dest = os.path.join(destinationfolder, name)
                isExist = os.path.exists(dest)
                if not isExist:
                    os.makedirs(dest)
                fname = name + "_" + str(z) + ".tif"
                fp = os.path.join(dest, fname)
                target_ds = driver.Create(fp, cols, rows, 1, gdal.GDT_CFloat32)
                target_ds.SetGeoTransform(gt)
                target_ds.SetProjection(crs)
                band = target_ds.GetRasterBand(1)
                band.WriteArray(data, 0, 0)
                band.FlushCache()
                res = None


def createInput2DFilesProject(rasterfile, inputfile, outputfile):
    ds = gdal.Open(rasterfile)
    driver = gdal.GetDriverByName('GTiff')
    gt = ds.GetGeoTransform()
    crs = ds.GetProjection()
    cols = ds.RasterXSize
    rows = ds.RasterYSize
    print("Processing Data: ", inputfile)
    data = np.load(inputfile)
    data = data.max(axis=0)
    target_ds = driver.Create(outputfile, cols, rows, 1, gdal.GDT_CFloat32)
    target_ds.SetGeoTransform(gt)
    target_ds.SetProjection(crs)
    band = target_ds.GetRasterBand(1)
    band.WriteArray(data, 0, 0)
    band.FlushCache()
    res = None


def createLithology2DFiles(rasterfile, lithologyfiles, outputfolder):
    ds = gdal.Open(rasterfile)
    driver = gdal.GetDriverByName('GTiff')
    gt = ds.GetGeoTransform()
    crs = ds.GetProjection()
    cols = ds.RasterXSize
    rows = ds.RasterYSize
    lt = []
    for filename in lithologyfiles:
        if ".npy" in filename:
            print("Processing Data: ", filename)
            array_data = np.load(filename)
            lt.append(array_data)

    elev = 850
    for z in tqdm(range(lt[0].shape[0])):
        res = np.zeros((lt[0].shape[1], lt[0].shape[2]))
        for i in range(len(lt)):
            arr3d = lt[i]
            arr2d = arr3d[z, :, :]
            res = res + ((i + 1) * arr2d)
        min = np.nanmin(res)
        max = np.nanmax(res)
        print(z, min, max)
        isExist = os.path.exists(outputfolder)
        if not isExist:
            os.makedirs(outputfolder)
        fname = "lithology_" + str(elev) + ".tif"
        elev = elev - 100
        fp = os.path.join(outputfolder, fname)
        target_ds = driver.Create(fp, cols, rows, 1, gdal.GDT_Int32)
        target_ds.SetGeoTransform(gt)
        target_ds.SetProjection(crs)
        band = target_ds.GetRasterBand(1)
        band.WriteArray(res, 0, 0)
        band.FlushCache()
        res = None


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument("-g", "--grid", required=False, help="flag to create grid shape file", action='store_true',
                    default=False)
    ap.add_argument("-v", "--voxel", required=False, help="flag to create 3D voxel ID csv file", action='store_true',
                    default=False)
    ap.add_argument("-m", "--map", required=False, help="flag to create 2D raster files", action='store_true',
                    default=False)
    ap.add_argument("-p", "--project", required=False, help="flag to create 2D raster projection file",
                    action='store_true', default=False)
    ap.add_argument("-l", "--lth", required=False, help="flag to create lithology files", action='store_true',
                    default=False)
    ap.add_argument("-s", "--source", required=False, help="source shapefile", default=None)
    ap.add_argument("-d", "--dest", required=False, help="destination shapefile", default=False)
    ap.add_argument("-x", "--xsource", required=False, help="Sources for X coordinates csv file", default=None)
    ap.add_argument("-y", "--ysource", required=False, help="Sources for Y coordinates csv file", default=None)
    ap.add_argument("-z", "--zsource", required=False, help="Sources for Z coordinates csv file", default=None)
    ap.add_argument("-sf", "--sourcefolder", required=False, help="source folder", default=None)
    ap.add_argument("-df", "--destfolder", required=False, help="destination folder", default=None)
    ap.add_argument("-i", "--input", required=False, help="input raster file", default=None)
    ap.add_argument("-t", "--lithology", required=False, help="help", nargs='+', default=[])
    args = vars(ap.parse_args())
    grl = args["grid"]
    vxl = args["voxel"]
    x = args["xsource"]
    y = args["ysource"]
    z = args["zsource"]
    src = args["source"]
    dst = args["dest"]
    map = args["map"]
    lth = args["lth"]
    prj = args["project"]
    sf = args["sourcefolder"]
    df = args["destfolder"]
    inp = args["input"]
    lths = args["lithology"]

    if grl and x is not None and y is not None and src is not None and dst is not None:
        create3DModelGrill(src, dst, x, y)
    elif vxl and src is not None and dst is not None and z is not None:
        write3DValuestoCSV(src, z, dst)
    elif map and src is not None and sf is not None and df is not None:
        createInput2DFiles(src, sf, df)
    elif prj and src is not None and inp is not None and dst is not None:
        createInput2DFilesProject(src, inp, dst)
    elif lth and src is not None and lths is not [] and df is not None:
        createLithology2DFiles(src, lths, df)
    else:
        print("wrong parameters")
