# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import os.path
import argparse
import constants as ct
import numpy as np
import pandas as pd


def addData(data, sline):
    x = sline.split()
    try:
        if data is not None:
            for value in x:
                data.append(float(value))
            return True
        print(sline)
        return False
    except ValueError:
        print(sline)
        return False
    except AttributeError:
        print(sline)
        return False


def getTextIndex(namelist, sline):
    try:
        return namelist.index(sline)
    except ValueError:
        return -1


def getData(index, datalist):
    if index >= 0:
        return datalist[index]
    else:
        return None


def printValues(namelist, datalist):
    for nm in namelist:
        # print(nm)
        ind = getTextIndex(namelist, nm)
        data = datalist[ind]
        u_data = np.unique(data)
        # print(u_data.max())
        # print(u_data.min())
        # print(np.percentile(data, q=[0, 10, 30, 50, 70, 90, 100]))
        print(nm, np.array2string(np.percentile(data, q=[0, 10, 30, 50, 70, 90, 100]), separator=" "))
        if len(u_data) < 6:
            print("Unique Values:", nm, u_data)


def getValues(name, namelist, datalist):
    ind = getTextIndex(namelist, name)
    data = datalist[ind]
    u_data = np.unique(data)
    if u_data.size <= 5:
        return u_data
    else:
        return np.percentile(data, q=[10, 30, 50, 70, 90])


def readAndSave(sourcefile, destinationfolder, print):
    co2015file = open(sourcefile, "r")

    data = None
    for line in co2015file:
        sline = line.strip()
        if not addData(data, sline):
            index = getTextIndex(ct.namelist, sline)
            data = getData(index, ct.datalist)

    # Closing files
    co2015file.close()
    if print:
        printValues(ct.namelist, ct.datalist)

    isExist = os.path.exists(destinationfolder)
    if not isExist:
        os.makedirs(destinationfolder)
    for name in ct.namelist:
        ind = getTextIndex(ct.namelist, name)
        data = getData(ind, ct.datalist)
        dt = np.array(data)
        filename = name + ".csv"
        fullpath = os.path.join(destinationfolder, filename)
        dict_val = {name: dt}
        df_val = pd.DataFrame(dict_val)
        df_val.to_csv(fullpath)


def readAndPrint(destination):
    for name in ct.namelist:
        filename = name + ".csv"
        fullpath = os.path.join(destination, filename)
        df = pd.read_csv(fullpath)
        ind = getTextIndex(ct.namelist, name)
        setData(ind, ct.datalist, np.array(df[name]))
    printValues(ct.namelist, ct.datalist)


def setData(index, datalist, val):
    if index >= 0:
        datalist[index] = val


def readAndSaveWellFile(source, dest1, dest2):
    currentwell = None
    welllist = []
    welllocations = []
    wellfile = open(source, "r")
    for line in wellfile:
        sline = line.strip()
        sline = sline.replace("'", "")
        sline = sline.replace(",", "")
        lines = sline.split()
        if len(lines) > 0:
            if lines[0] == ct.P:
                type = ct.P
                name = lines[1]
                currentwell = []
                currentwell.append(name)
                currentwell.append(type)
                welllocations = []
                currentwell.append(welllocations)
                welllist.append(currentwell)
            elif lines[0] == ct.I:
                type = ct.I
                name = lines[1]
                currentwell = []
                currentwell.append(name)
                currentwell.append(type)
                welllocations = []
                currentwell.append(welllocations)
                welllist.append(currentwell)
            elif currentwell is not None:
                for ln in lines:
                    try:
                        loc = int(ln)
                        welllocations.append(loc)
                    except ValueError:
                        print(ln)
                currentwell = None
    wl = pd.DataFrame(welllist, columns=["Name", "Type", "Location"])
    wl.to_csv(dest1)
    data = np.zeros(35904)
    for well in welllist:
        # print(well[0], well[1], well[2])
        if well[1] == ct.P:
            wl = np.array(well[2])
            for loc in wl:
                data[loc - 1] = 1
        if well[1] == ct.I:
            wl = np.array(well[2])
            for loc in wl:
                data[loc - 1] = 2
    pd.DataFrame(data, columns=["val"]).to_csv(dest2)
    print("DONE!")


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument("-w", "--well", required=False, help="Source file to import", action='store_true', default=None)
    ap.add_argument("-s", "--source", required=False, help="Source file to import", default=None)
    ap.add_argument("-d", "--destination", required=False, help="Destination folder to write csv files", default=None)
    ap.add_argument("-f1", "--file1", required=False, help="Destination file to write csv file", default=None)
    ap.add_argument("-f2", "--file2", required=False, help="Destination file to write csv file", default=None)
    ap.add_argument("-p", "--print", required=False, help="print ", action='store_true',
                    default=False)
    args = vars(ap.parse_args())
    well = args["well"]
    source = args["source"]
    destination = args["destination"]
    file1 = args["file1"]
    file2 = args["file2"]
    prnt = args["print"]
    if well and file1 is not None and file2 is not None:
        readAndSaveWellFile(source, file1, file2)
    elif source is not None and destination is not None:
        readAndSave(source, destination, prnt)
    elif destination is not None:
        readAndPrint(destination)
    else:
        print("wrong parameters")
