# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import math

import numpy as np
import pandas as pd
from numpy.linalg import LinAlgError

import constants
import constants as ct
import argparse
import os

from osgeo import gdal, ogr, osr
from tqdm import tqdm


def createInputDataCont(rasterfile, inputcsv, inputcsvcol, resultfile, zlenght=12300):
    exp = pd.read_csv(inputcsv)
    print(exp.head())
    print(exp.tail())
    print(exp.shape)
    ex = exp[["X", "Y", "Z"]]
    print(ex.head())
    for col in ex:
        print(exp[col].unique().size)
        print(exp[col].unique())
    ds = gdal.Open(rasterfile)
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    gt = ds.GetGeoTransform()
    xref = gt[0]
    yref = gt[3]
    zref = ct.Z_REF
    size = gt[1]
    center = size / 2
    zlenght = zlenght * ct.FEET
    zsize = int(zlenght / size) + 1
    result = np.zeros((zsize, ysize, xsize))
    result[:] = np.nan
    for z in tqdm(range(result.shape[0])):
        zmin = zref - (z * size) - center
        zmax = zref - (z * size) + center
        zls = exp[(exp["Z"] <= zmax) & (exp["Z"] > zmin)]
        for y in tqdm(range(result.shape[1])):
            ymin = yref - (y * size) - center
            ymax = yref - (y * size) + center
            yls = zls[(zls["Y"] < ymax) & (zls["Y"] >= ymin)]
            for x in range(result.shape[2]):
                xmin = xref + (x * size) - center
                xmax = xref + (x * size) + center
                xls = yls[(yls["X"] <= xmax) & (yls["X"] > xmin)]
                if xls.shape[0] == 1:
                    val = xls.iloc[0]
                    result[z][y][x] = getVal(val, inputcsvcol)
                else:
                    result[z][y][x] = np.nan
    np.save(resultfile, result)


def createGroundTruthData(rasterfile, inputcsv, inputcsvcol, threshold, voxelidcsv, voxelidcsvcol, resultfile,
                          zlenght=12300):
    thr = float(threshold)
    input = pd.read_csv(inputcsv)
    input = input[inputcsvcol]
    inputarray = input.to_numpy()
    vxl = pd.read_csv(voxelidcsv)
    # vxlarr = np.array(vxl["X", "Y", "Z", voxelidcsvcol])
    vxlarr = vxl.to_numpy()
    ds = gdal.Open(rasterfile)
    gt = ds.GetGeoTransform()
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    print(xsize)
    print(ysize)
    xref = gt[0]
    yref = gt[3]
    zref = ct.Z_REF
    size = gt[1]
    center = size / 2
    zlenght = zlenght * ct.FEET
    zsize = int(zlenght / size) + 1
    result = np.zeros((zsize, ysize, xsize))
    for z in tqdm(range(result.shape[0])):
        for y in tqdm(range(result.shape[1])):
            for x in tqdm(range(result.shape[2])):
                diff = vxlarr - np.array([0, xref + (x * size), yref - (y * size), zref - (z * size), 0])
                srt = diff[:, 1:4]
                srt = np.abs(srt)
                dst = np.linalg.norm(srt, axis=1)
                minn = np.argmin(dst, axis=0)
                xdst = srt[minn, 0]
                ydst = srt[minn, 1]
                zdst = srt[minn, 2]
                vid = diff[minn, 4]
                if xdst <= size and ydst <= size and zdst <= size and vid != ct.NO_DATA:
                    res = inputarray[int(vid) - 1]
                    if res >= thr:
                        result[z][y][x] = 1
                    else:
                        result[z][y][x] = 0
                else:
                    result[z][y][x] = np.nan
                # print(x, y, z)
    np.save(resultfile, result)


def createGroundTruthData2(rasterfile, inputcsv, inputcsvcol, threshold, resultfile, thicknessfile, zlenght=12300):
    thr = float(threshold)
    input = pd.read_csv(inputcsv)
    input = input[inputcsvcol]
    inputarray = input.to_numpy()
    inputarray = inputarray.reshape(-1, 2112)
    ds = gdal.Open(rasterfile)
    band = ds.GetRasterBand(1)
    data = band.ReadAsArray()
    z_thick = pd.read_csv(thicknessfile)
    zarray = np.array(z_thick["Thickness"]).reshape(-1, 2112)
    zl = zarray[:, 0]
    gt = ds.GetGeoTransform()
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    print(xsize)
    print(ysize)
    size = gt[1]
    zlenght = zlenght * ct.FEET
    zsize = int(zlenght / size) + 1
    result = np.zeros((zsize, ysize, xsize)) * -1
    for z in tqdm(range(result.shape[0])):
        d = np.array(data, dtype=int, copy=True)
        d[data == 65535] = -1
        zid = getVoxelZ(z, zl, size)
        val = inputarray[zid,]
        for vxid in range(2112):
            if val[vxid] >= thr:
                d[data == (vxid + 1)] = 1
            else:
                d[data == (vxid + 1)] = 0
        result[z, :, :] = d
    np.save(resultfile, result)


def write3DValuestoNPY(sourcerasterfile, thicknessfile, inputcsv, inputcsvcol, threshold, resultfile):
    ds = gdal.Open(sourcerasterfile)
    z_thick = pd.read_csv(thicknessfile)
    incsv = pd.read_csv(inputcsv)
    incsv = incsv[inputcsvcol]
    inputarray = incsv.to_numpy()
    zarray = np.array(z_thick["Thickness"]).reshape(-1, 2112)
    zl = zarray[:, 0]
    gt = ds.GetGeoTransform()
    data = ds.GetRasterBand(1)
    dt = data.ReadAsArray()
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    size = gt[1]
    zlenght = 12300
    zlenght = zlenght * ct.FEET
    zsize = int(zlenght / size) + 1
    res = np.zeros((zsize, ysize, xsize))
    for z in tqdm(range(zsize)):
        cp = np.array(dt.copy(), dtype=float)
        cp[cp == 65535] = np.nan
        for v in range(2112):
            zp = getVoxelZ(z, zl, size)
            vl = v + (zp * 2112)
            cp[cp == (v + 1)] = inputarray[vl]
        res[z, :, :] = cp
    rs = res >= threshold
    np.save(resultfile, rs)


def getVoxelZ(step, zl, size):
    total = 0
    for i in range(len(zl)):
        total = total + zl[i]
        if (step * size) < (total * ct.FEET):
            return i


def createInputDataCategorical(rasterfile, inputcsv, inputcsvcol, resultfolder, zlenght=12300):
    exp = pd.read_csv(inputcsv)
    ds = gdal.Open(rasterfile)
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    gt = ds.GetGeoTransform()
    xref = gt[0]
    yref = gt[3]
    zref = ct.Z_REF
    size = gt[1]
    center = size / 2
    zlenght = zlenght * ct.FEET
    zsize = int(zlenght / size) + 1
    categories = exp[inputcsvcol].unique()
    data = []

    exp["CategoryIndex"] = -1
    for cat in categories:
        if isinstance(cat, str):
            ind = np.where(categories == cat)
            exp.loc[exp[inputcsvcol] == cat, "CategoryIndex"] = ind

    exp = exp[["X", "Y", "Z", "CategoryIndex"]]
    res = np.zeros((zsize, ysize, xsize)) * (-1)

    for z in tqdm(range(res.shape[0])):
        zmin = zref - (z * size) - center
        zmax = zref - (z * size) + center
        zls = exp[(exp["Z"] <= zmax) & (exp["Z"] > zmin)]
        for y in tqdm(range(res.shape[1])):
            ymin = yref - (y * size) - center
            ymax = yref - (y * size) + center
            yls = zls[(zls["Y"] < ymax) & (zls["Y"] >= ymin)]
            for x in range(res.shape[2]):
                xmin = xref + (x * size) - center
                xmax = xref + (x * size) + center
                xls = yls[(yls["X"] <= xmax) & (yls["X"] > xmin)]
                if xls.shape[0] == 1:
                    val = xls.iloc[0]
                    res[z][y][x] = val["CategoryIndex"]

    values, counts = np.unique(res, return_counts=True)
    print(values)
    print(counts)
    for cat in categories:
        if isinstance(cat, str):
            ind = np.where(categories == cat)
            rs = res.copy()
            rs[rs != ind] = -1
            rs[rs == ind] = 1
            rs[rs == -1] = 0
            values, counts = np.unique(rs, return_counts=True)
            print(values)
            print(counts)
            lcat = cat.strip().replace(" ", "").lower()
            name = lcat + ".npy"
            fullpath = os.path.join(resultfolder, name)
            np.save(fullpath, rs)


def getVal(row, column):
    if isinstance(row[column], float):
        return row[column]
    elif row[column] != "OUT":
        return float(row[column])
    else:
        return np.nan


def createFaultDensity(rasterfile, faultfolder, resultfile, zlenght=12300):
    faults = []
    for filename in os.listdir(faultfolder):
        faults.append(np.asarray(pd.read_csv(os.path.join(faultfolder, filename))))
    ds = gdal.Open(rasterfile)
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    gt = ds.GetGeoTransform()
    xref = gt[0]
    yref = gt[3]
    zref = ct.Z_REF
    size = gt[1]
    center = size / 2
    zlenght = zlenght * ct.FEET
    zsize = int(zlenght / size) + 1
    result = np.zeros((zsize, ysize, xsize))
    for z in tqdm(range(result.shape[0])):
        zc = zref - (z * size)
        zmin = zref - (z * size) - center
        zmax = zref - (z * size) + center
        for y in tqdm(range(result.shape[1])):
            yc = yref - (y * size)
            ymin = yref - (y * size) - center
            ymax = yref - (y * size) + center
            for x in range(result.shape[2]):
                xc = xref + (x * size)
                center_c = np.array([xc, yc, zc])
                xmin = xref + (x * size) - center
                xmax = xref + (x * size) + center
                for fault in faults:
                    dx = np.abs(fault[:, 0] - xc)
                    dx2 = dx * dx
                    dy = np.abs(fault[:, 1] - yc)
                    dy2 = dy * dy
                    dz = np.abs(fault[:, 2] - zc)
                    dz2 = dz * dz
                    dt = dx2 + dy2 + dz2
                    dt = np.sqrt(dt)
                    id = np.argpartition(dt, 4)
                    x1, y1, z1 = fault[id[0], 0], fault[id[0], 1], fault[id[0], 2]
                    x2, y2, z2 = fault[id[1], 0], fault[id[1], 1], fault[id[1], 2]
                    x3, y3, z3 = fault[id[2], 0], fault[id[2], 1], fault[id[2], 2]
                    # x4, y4, z4 = fault[id[3], 0], fault[id[3], 1], fault[id[3], 2]
                    x1a = np.array([x1, y1, z1])
                    x2a = np.array([x2, y2, z2])
                    x3a = np.array([x3, y3, z3])
                    r = compareTriangleAndCube(x1a, x2a, x3a, xmin, xmax, ymin, ymax, zmin, zmax, center_c)
                    if r is None:
                        print("r value is None")
                    if r:
                        result[z][y][x] = result[z][y][x] + 1
    unique, counts = np.unique(result, return_counts=True)
    print(unique, counts)
    np.save(resultfile, result)


def compareTriangleAndCube(a, b, c, xmin, xmax, ymin, ymax, zmin, zmax, center_c):
    # 3^9 cases exist
    ax, ay, az = isInCubeCompare(a, xmin, xmax, ymin, ymax, zmin, zmax)
    bx, by, bz = isInCubeCompare(b, xmin, xmax, ymin, ymax, zmin, zmax)
    cx, cy, cz = isInCubeCompare(c, xmin, xmax, ymin, ymax, zmin, zmax)
    # 3^6 cases
    if ax == 1 and bx == 1 and cx == 1:
        return False
    # 3^6 cases
    if ax == -1 and bx == -1 and cx == -1:
        return False
    # 3^6 cases
    if ay == 1 and by == 1 and cy == 1:
        return False
    # 3^6 cases
    if ay == -1 and by == -1 and cy == -1:
        return False
    # 3^6 cases
    if az == 1 and bz == 1 and cz == 1:
        return False
    # 3^6 cases
    if az == -1 and bz == -1 and cz == -1:
        return False
    # 3 * 3^6 = 3^7 cases
    if (ax == 0 and ay == 0 and az == 0) or (bx == 0 and by == 0 and bz == 0) or (cx == 0 and cy == 0 and cz == 0):
        return True
    # 2 * 3^4 cases
    if lineCrossing(ax, ay, az, bx, by, bz):
        return True
    # 2 * 3^4 cases
    if lineCrossing(ax, ay, az, cx, cy, cz):
        return True
    # 2 * 3^4 cases
    if lineCrossing(cx, cy, cz, bx, by, bz):
        return True
    # get mid-point and compare which points are closest
    avg = (a + b + c) / 3
    mt = np.array([a, b, c, avg])
    mtt = np.array([a - center_c, b - center_c, c - center_c, avg - center_c])
    dt = np.linalg.norm(mtt, axis=1)
    id = np.argpartition(dt, 3)
    aa = mt[id[0], :]
    bb = mt[id[1], :]
    cc = mt[id[2], :]
    return compareTriangleAndCube(aa, bb, cc, xmin, xmax, ymin, ymax, zmin, zmax, center_c)


# 2 * 3^4 cases
def lineCrossing(ax, ay, az, bx, by, bz):
    # 3^3 cases
    if ax == 0 and ay == 0 and az == 1 and bx == 0 and by == 0 and bz == -1:
        return True
    # 3^3 cases
    if ax == 0 and ay == 0 and az == -1 and bx == 0 and by == 0 and bz == 1:
        return True
    # 3^3 cases
    if ax == 0 and ay == 1 and az == 0 and bx == 0 and by == -1 and bz == 0:
        return True
    # 3^3 cases
    if ax == 0 and ay == -1 and az == 0 and bx == 0 and by == 1 and bz == 0:
        return True
    # 3^3 cases
    if ax == 1 and ay == 0 and az == 0 and bx == -1 and by == 0 and bz == 0:
        return True
    # 3^3 cases
    if ax == -1 and ay == 0 and az == 0 and bx == 1 and by == 0 and bz == 0:
        return True


def isInCubeCompare(p, xmin, xmax, ymin, ymax, zmin, zmax):
    xval = 0
    if p[0] > xmax:
        xval = 1
    elif p[0] < xmin:
        xval = -1
    yval = 0
    if p[1] > ymax:
        yval = 1
    elif p[1] < ymin:
        yval = -1
    zval = 0
    if p[2] > zmax:
        zval = 1
    elif p[2] < zmin:
        zval = -1
    return xval, yval, zval


def getmask(radius):
    radsquare = radius * radius
    rad = int(radius) + 1
    kernelsize = 2 * rad + 1
    mask = np.full((kernelsize, kernelsize, kernelsize), True, dtype=bool)
    for x in range(2 * rad + 1):
        for y in range(2 * rad + 1):
            for z in range(2 * rad + 1):
                dist = ((x - rad) * (x - rad)) + ((y - rad) * (y - rad)) + ((z - rad) * (z - rad))
                if dist < radsquare:
                    mask[z, y, x] = True
                else:
                    mask[z, y, x] = False
    return mask, kernelsize, rad


def createFaultDensity2(source, dest, radius, pixel_size):
    res = np.load(source)
    result = np.zeros((res.shape[0], res.shape[1], res.shape[2]))
    rad = radius / pixel_size
    mask, kernelsize, rad = getmask(rad)
    p_data = np.pad(res, (rad,), 'constant', constant_values=0)
    for z in tqdm(range(result.shape[0])):
        for y in tqdm(range(result.shape[1])):
            for x in tqdm(range(result.shape[2])):
                val = p_data[z:z + kernelsize, y:y + kernelsize, x:x + kernelsize]
                mval = np.ma.masked_array(val, mask=mask)
                tot = np.nansum(mval)
                result[z][y][x] = tot

    unique, counts = np.unique(result, return_counts=True)
    print(np.array([unique, counts], dtype=np.int32).T)
    np.save(dest, result)


def createFaultDensity3(source, dest, radius, pixel_size):
    res = np.load(source)
    res = np.nan_to_num(res)
    x_size = res.shape[2]
    y_size = res.shape[1]
    z_size = res.shape[0]
    result = np.zeros((z_size, y_size, x_size))
    rad = radius / pixel_size
    mskk, kernelsize, rad = getmask(rad)
    indexes = np.array(np.asarray(mskk == True).nonzero())
    p_data = np.pad(res, (rad,), 'constant', constant_values=0)
    for ind in tqdm(range(indexes.shape[1])):
        z_shift = indexes[0, ind]
        y_shift = indexes[1, ind]
        x_shift = indexes[2, ind]
        val = p_data[z_shift:z_shift + z_size, y_shift:y_shift + y_size, x_shift:x_shift + x_size]
        result = result + val
    unique, counts = np.unique(result, return_counts=True)
    print(np.array([unique, counts], dtype=np.int32).T)
    np.save(dest, result)


def createInputDataCategorical2(rasterfile, inputcsv, inputcsvcol, valfile, neigfile, zlenght=12300):
    exp = pd.read_csv(inputcsv)
    ds = gdal.Open(rasterfile)
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    gt = ds.GetGeoTransform()
    xref = gt[0]
    yref = gt[3]
    zref = ct.Z_REF
    size = gt[1]
    center = size / 2
    zlenght = zlenght * ct.FEET
    zsize = int(zlenght / size) + 1
    categories = exp[inputcsvcol].unique()
    print(categories)
    val = []
    neig = []
    for z in tqdm(range(zsize)):
        for y in tqdm(range(ysize)):
            for x in range(xsize):
                xmin = xref + (x * size) - center
                xmax = xref + (x * size) + center
                ymin = yref - (y * size) - center
                ymax = yref - (y * size) + center
                zmin = zref - (z * size) - center
                zmax = zref - (z * size) + center
                ls = exp[(exp["X"] <= xmax) & (exp["X"] > xmin) & (exp["Y"] < ymax) & (exp["Y"] >= ymin) &
                         (exp["Z"] <= zmax) & (exp["Z"] > zmin)]
                if ls.shape[0] == 1:
                    vl = ls.iloc[0]
                    lit = vl[inputcsvcol]
                    val.append(constants.lithology.index(lit))
                    neig.append(getneigh(x, y, z, xsize, ysize, zsize))
                    id = (z * (ysize * xsize)) + (y * xsize) + (x + 1)
                    # print(id)
                elif ls.shape[0] == 0:
                    lit = -1
                    val.append(lit)
                    neig.append(getneigh(x, y, z, xsize, ysize, zsize))
                    id = (z * (ysize * xsize)) + (y * xsize) + (x + 1)
                    # print(id)
                else:
                    print("something wrong")
    pd.DataFrame(val, columns=["category"]).to_csv(valfile)
    pd.DataFrame(neig, columns=["neighbour"]).to_csv(neigfile)


def getneigh(x, y, z, xsize, ysize, zsize):
    # id = (z * (ysize * xsize)) + (y * xsize) + (x + 1)
    n1x = (z * ysize * xsize) + (y * xsize) + (x)
    n2x = (z * ysize * xsize) + (y * xsize) + (x + 2)
    n1y = (z * ysize * xsize) + ((y - 1) * xsize) + (x + 1)
    n2y = (z * ysize * xsize) + ((y + 1) * xsize) + (x + 1)
    n1z = ((z - 1) * ysize * xsize) + (y * xsize) + (x + 1)
    n2z = ((z + 1) * ysize * xsize) + (y * xsize) + (x + 1)

    if x == 0: n1x = 0
    if x == (xsize - 1): n2x = 0
    if y == 0: n1y = 0
    if y == (ysize - 1): n2y = 0
    if z == 0: n1z = 0
    if z == (zsize - 1): n2z = 0

    res = ""
    if n1x > 0:
        res = res + "," + str(n1x)
    if n2x > 0:
        res = res + "," + str(n2x)
    if n1y > 0:
        res = res + "," + str(n1y)
    if n2y > 0:
        res = res + "," + str(n2y)
    if n1z > 0:
        res = res + "," + str(n1z)
    if n2z > 0:
        res = res + "," + str(n2z)
    return res


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument("-y", "--faultdensity", required=False, help="flag to create fault file", action='store_true',
                    default=False)
    ap.add_argument("-z", "--fault", required=False, help="flag to create fault file", action='store_true',
                    default=False)
    ap.add_argument("-g", "--ground", required=False, help="flag to create ground truth", action='store_true',
                    default=False)
    ap.add_argument("-v", "--voxelid", required=False, help="voxel ID csv file", default=False)
    ap.add_argument("-r", "--raster", required=False, help="source rasterfile", default=None)
    ap.add_argument("-i", "--input", required=False, help="input csv file", default=False)
    ap.add_argument("-t", "--threshold", required=False, help="threshold value", default=None)
    ap.add_argument("-v2", "--voxelidcol", required=False, help="ID column of voxel ID csv file", default=None)
    ap.add_argument("-i2", "--inputcol", required=False, help="value column of input csv file", default=None)
    ap.add_argument("-s", "--result", required=False, help="result file (.npy)", default=None)
    ap.add_argument("-c", "--cont", required=False, help="flag to create continuos data", action='store_true',
                    default=False)
    ap.add_argument("-d", "--category", required=False, help="flag to create categorical data", action='store_true',
                    default=False)
    ap.add_argument("-f", "--folder", required=False, help="result folder", default=None)
    ap.add_argument("-p", "--pixelsize", required=False, help="result folder", default=0)
    ap.add_argument("-a", "--radius", required=False, help="result folder", default=0)
    ap.add_argument("-e", "--thickness", required=False, help="thickness csv file", default=None)

    args = vars(ap.parse_args())
    gnd = args["ground"]
    flt = args["fault"]
    cnt = args["cont"]
    cat = args["category"]
    vxl = args["voxelid"]
    rst = args["raster"]
    inp = args["input"]
    threshold = args["threshold"]
    voxelcolname = args["voxelidcol"]
    inputcolname = args["inputcol"]
    result = args["result"]
    folder = args["folder"]
    dns = args["faultdensity"]
    thic = args["thickness"]
    rd = int(args["radius"])
    px = int(args["pixelsize"])

    if gnd and thic is not None and rst is not None and inp is not None and threshold is not None and inputcolname is not None:
        createGroundTruthData2(rst, inp, inputcolname, threshold, result, thic)
    elif cnt and rst is not None and inp is not None and inputcolname is not None and result is not None:
        createInputDataCont(rst, inp, inputcolname, result)
    elif cat and rst is not None and inp is not None and inputcolname is not None and folder is not None:
        createInputDataCategorical(rst, inp, inputcolname, folder)
    elif flt and rst is not None and result is not None and folder is not None:
        createFaultDensity(rst, folder, result)
    elif dns and inp is not None and result is not None and rd is not None and px is not None:
        createFaultDensity2(inp, result, rd, px)
    else:
        print("wrong parameters")
