# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import pickle
import argparse
import os

from sklearn.preprocessing import LabelBinarizer
from tqdm import tqdm
import tensorflow as tf
import keras
from keras.callbacks import EarlyStopping, ModelCheckpoint
from keras.layers import AveragePooling3D, Input, Concatenate
from keras.layers import Conv3D, MaxPooling3D
from keras.layers import Activation, Flatten, Dropout, Dense
from keras.models import Model
from tensorflow.keras.optimizers.experimental import Adam, Nadam, Adadelta, Adagrad, Adamax, SGD
from keras.regularizers import l2
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report, confusion_matrix, accuracy_score
from numpy import argmax


class DataGenerator(keras.utils.Sequence):
    'Generates data for Keras'

    def __init__(self, data_set, labels, batch_size=200,
                 dim=(19, 19, 19, 3),
                 n_channels=3, n_classes=2,
                 shuffle=True, augment_data=False):
        'Initialization'
        self.dim = dim
        self.batch_size = batch_size
        self.labels = labels
        self.data_set = data_set
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.shuffle = shuffle
        self.augment_data = augment_data
        self.on_epoch_end()

    def __len__(self):
        'Denotes the number of batches per epoch'
        return int(np.floor(len(self.data_set) / self.batch_size))

    def __getitem__(self, index):
        'Generate one batch of data'
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size:(index + 1) * self.batch_size]
        # Generate data
        X, y = self.__data_generation(indexes)

        return X, y

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        self.indexes = np.arange(len(self.data_set))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __data_generation(self, list_IDs_temp):
        'Generates data containing batch_size samples'  # X : (n_samples, *dim, n_channels)
        # Initialization
        X = np.empty((self.batch_size, self.n_channels, *self.dim))
        y = np.empty((self.batch_size, self.n_classes), dtype=int)
        # Generate data
        for i, ID in enumerate(list_IDs_temp):
            # Store sample
            image_raw = self.data_set[ID]
            if self.augment_data:
                if np.random.randint(1):
                    image_raw = np.rot90(image_raw, 2)
                if np.random.randint(1):
                    image_raw = np.fliplr(image_raw)
                if np.random.randint(1):
                    image_raw = np.flipud(image_raw)
            X[i,] = image_raw
            # Store class
            y[i] = self.labels[ID]
        return X, y


def getData(layers, directory, kernel_pixels):
    padding = int(kernel_pixels / 2)
    data = []
    for layer in layers:
        ly = os.path.join(directory, layer)
        npa = np.load(ly)
        npa = np.pad(npa, (padding,), 'edge')
        npa = np.array(npa, dtype=np.float32)
        data.append(npa)
    return data


def getDataSelection(non_zero, directory, nosamples, gnd_lyr, zero_label, class_label, list):
    fl = os.path.join(directory, non_zero)
    nz = np.load(fl)
    gn = np.load(gnd_lyr)
    nz[(nz == zero_label) & (gn == class_label)] = 2
    n_z = np.array(np.where(nz == 2))
    sampleSize = int(nosamples)
    if sampleSize > len(n_z.T):
        sampleSize = len(n_z.T)
    print(non_zero, len(n_z.T), int(nosamples), sampleSize, zero_label, class_label)
    choices = np.random.choice(len(n_z.T), sampleSize, replace=False)
    for c in tqdm(choices, desc="Calculating...", ascii=False, ncols=75):
        z = n_z[0][c]
        y = n_z[1][c]
        x = n_z[2][c]
        val = [x, y, z]
        list.append(val)
    return list


def getSelectedData(data, list, gnd_layer, kernel_pixels):
    dt = []
    lbl = []
    gnd = np.load(gnd_layer)
    for c in tqdm(list):
        x = c[0]
        y = c[1]
        z = c[2]
        batch = []
        for band in data:
            batch.append(np.copy(band[z:z + kernel_pixels, y:y + kernel_pixels, x:x + kernel_pixels]))
        vl = np.array(batch)
        if not np.isnan(vl).any():
            dt.append(vl)
            lbl.append(gnd[z, y, x])
    alt_dt = np.asarray(dt, dtype=np.float32)
    labels = np.array(lbl)
    return alt_dt, labels


def createModel(number_of_bands, data, labels, kernel_pixels, batch_size, num_epochs, modelfile):
    if os.path.isfile(modelfile):
        print(modelfile, "exists. Exiting...")
        return
    print("[INFO] data matrix: {:.2f}MB".format(
        data.nbytes / (1024 * 1024.0)))
    # binarize the labels
    lb = LabelBinarizer()
    y_binary = tf.keras.utils.to_categorical(labels, num_classes=2)
    y_inverse = np.argmax(y_binary, axis=1)
    # print('Read labels:', (lb.classes_))
    # print('Transformed labels:', (y_binary[:10]))
    # print('Inverted labels:', (y_inverse[:10]))
    # print('Any nulls?: ', np.isnan(data).any())

    gpus = tf.config.list_physical_devices('GPU')
    if gpus:
        try:
            # Currently, memory growth needs to be the same across GPUs
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.list_logical_devices('GPU')
            print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Memory growth must be set before GPUs have been initialized
            print(e)
    ###
    '''
    Creates the network
    '''
    # Define distribution strategy
    strategy = tf.distribute.MirroredStrategy()
    # construct model under distribution strategy scope
    with strategy.scope():
        my_input = Input(shape=(number_of_bands, kernel_pixels, kernel_pixels, kernel_pixels), name="input")
        # One jigsaw module(s)
        jigsaw_01 = jigsaw_m(my_input)
        # Attaches end to jigsaw modules, returns class within num_classes
        loss3_classifier_act = jigsaw_m_end(jigsaw_01, num_classes=2, first_layer=my_input)  # testing num_classes
        # Builds model
        model3 = Model(inputs=my_input, outputs=[loss3_classifier_act])
        model3.summary()

        # (trainX, testX, trainY, testY) = train_test_split(data, y_binary, test_size=0.8, random_state=42)
        # (validateX, testX, validateY, testY) = train_test_split(testX, testY, test_size=0.75, random_state=42)
        params = {'dim': (kernel_pixels, kernel_pixels, kernel_pixels), 'batch_size': batch_size, 'n_classes': 2,
                  'n_channels': number_of_bands, 'shuffle': True, 'augment_data': False}
        # construct the image generator for data augmentation
        # print("TrainX", trainX.shape)
        # print("TrainY", trainY.shape)
        # my_batch_gen = DataGenerator(trainX, trainY, **params)
        # print('creating generator with trainX, trainY of shapes: (%s, %s)' %
        #      (trainX.shape, trainY.shape)
        #      )
        print("TrainX", data.shape)
        print("TrainY", y_binary.shape)
        my_batch_gen = DataGenerator(data, y_binary, **params)
        print('creating generator with trainX, trainY of shapes: (%s, %s)' %
              (data.shape, y_binary.shape)
              )
        ## initialize the model
        print("[INFO] compiling model...")
        print("[INFO] Training...")
        # Hyperparameter selection: Optimization function
        # opt = Adam(lr=INIT_LR, decay=INIT_DECAY)   # Old decay was: INIT_LR / EPOCHS)
        # opt = Adam(lr=INIT_LR, beta_1=INIT_DECAY, amsgrad=True)
        # opt = Adadelta(learning_rate=INIT_LR)
        opt = Adadelta()
        model3.compile(loss="categorical_crossentropy", optimizer=opt, metrics=["accuracy"])
        # define the network's early stopping
        print("[INFO] define early stop and auto save for network...")
        auto_save = ModelCheckpoint(modelfile, monitor='accuracy', verbose=0,
                                    save_best_only=True, save_weights_only=False,
                                    mode='auto')
        # can use validation set loss or accuracy to stop early
        early_stop = EarlyStopping(monitor='accuracy', mode='max', baseline=0.97)
        # patience was 50
        # early_stop = EarlyStopping(monitor='val_loss', mode='min', verbose=1, patience=10)
        # train the network
        print("[INFO] training network...")
        # Train the model
        H = model3.fit(
            my_batch_gen,
            # validation_data=(validateX, validateY),
            # steps_per_epoch=len(trainX) // batch_size,
            steps_per_epoch=len(data) // batch_size,
            # callbacks=[auto_save, early_stop],
            callbacks=[auto_save],
            epochs=num_epochs, verbose=1)
        # save the model to disk
        print("[INFO] Saving model file...", modelfile)
        model3.save(modelfile)


def jigsaw_m(input_net, first_layer=None):
    conv1 = Conv3D(128, (1, 1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002), name="conv1")(
        input_net)
    jigsaw_t1_1x1 = Conv3D(256, (1, 1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                           name="jigsaw_t1_1x1")(conv1)
    jigsaw_t1_3x3_reduce = Conv3D(96, (1, 1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                                  name="jigsaw_t1_3x3_reduce")(conv1)
    jigsaw_t1_3x3 = Conv3D(128, (3, 3, 3), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                           name="jigsaw_t1_3x3")(jigsaw_t1_3x3_reduce)
    jigsaw_t1_5x5_reduce = Conv3D(16, (1, 1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                                  name="jigsaw_t1_5x5_reduce")(conv1)
    jigsaw_t1_5x5 = Conv3D(128, (5, 5, 5), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                           name="jigsaw_t1_5x5")(jigsaw_t1_5x5_reduce)
    jigsaw_t1_7x7_reduce = Conv3D(16, (1, 1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                                  name="jigsaw_t1_7x7_reduce")(conv1)
    jigsaw_t1_7x7 = Conv3D(128, (7, 7, 7), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                           name="jigsaw_t1_7x7")(jigsaw_t1_7x7_reduce)
    jigsaw_t1_9x9_reduce = Conv3D(16, (1, 1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                                  name="jigsaw_t1_9x9_reduce")(conv1)
    jigsaw_t1_9x9 = Conv3D(64, (9, 9, 9), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                           name="jigsaw_t1_9x9")(jigsaw_t1_9x9_reduce)
    jigsaw_t1_pool = MaxPooling3D(pool_size=(3, 3, 3), strides=(1, 1, 1), padding='same', name="jigsaw_t1_pool")(conv1)
    jigsaw_t1_pool_proj = Conv3D(32, (1, 1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                                 name="jigsaw_t1_pool_proj")(jigsaw_t1_pool)
    if first_layer is None:
        jigsaw_t1_output = Concatenate(axis=-1, name="jigsaw_t1_output")([jigsaw_t1_1x1, jigsaw_t1_3x3, jigsaw_t1_5x5,
                                                                          jigsaw_t1_7x7, jigsaw_t1_9x9,
                                                                          jigsaw_t1_pool_proj])
    else:
        jigsaw_t1_first = Conv3D(96, (1, 1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                                 name="jigsaw_t1_output")(first_layer)
        jigsaw_t1_output = Concatenate(axis=-1, name="jigsaw_t1_output")([jigsaw_t1_first, jigsaw_t1_1x1, jigsaw_t1_3x3,
                                                                          jigsaw_t1_5x5, jigsaw_t1_7x7, jigsaw_t1_9x9,
                                                                          jigsaw_t1_pool_proj])
    return jigsaw_t1_output


def jigsaw_m_end(input_net, num_classes=2, first_layer=None):
    avg_pooling = AveragePooling3D(pool_size=(3, 3, 3), padding='same', strides=(1, 1, 1), name='avg_pooling')(
        input_net)
    flat = Flatten(name="flat")(avg_pooling)
    flat = Dense(16, kernel_regularizer=l2(0.002), name="dense16")(flat)
    flat = Dropout(0.4, name="droput")(flat)
    if first_layer is not None:
        input_pixel = Flatten()(first_layer)
        input_pixel = Dense(16, kernel_regularizer=l2(0.002))(input_pixel)
        input_pixel = Dropout(0.2)(input_pixel)
        input_pixel = Dense(16, kernel_regularizer=l2(0.002))(input_pixel)
        input_pixel = Dropout(0.2)(input_pixel)
        flat = Concatenate(axis=-1, name="concat1")([input_pixel, flat])
    flat = Dense(32, kernel_regularizer=l2(0.002), name="dense32")(flat)
    avg_pooling = Dropout(0.4)(flat)
    print('num classes: ', num_classes)
    loss3_classifier = Dense(num_classes, kernel_regularizer=l2(0.002), name="denselc")(avg_pooling)
    loss3_classifier_act = Activation('softmax', name='prob')(loss3_classifier)
    return loss3_classifier_act


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument("-l", "--layers", required=True, help="input layers", nargs='+', default=[])
    ap.add_argument("-d", "--dir", required=True, help="ground truth", default=None)
    ap.add_argument("-n", "--nonzero", required=True, help="ground truth", default=None)
    ap.add_argument("-g", "--ground", required=True, help="ground truth", default=None)
    ap.add_argument("-k", "--kernelpixels", required=True, help="kernel pixels size", default=0)
    ap.add_argument("-s", "--sample", required=True, help="positive sample size", nargs='+', default=[])
    ap.add_argument("-b", "--batch", required=True, help="batch size", default=0)
    ap.add_argument("-e", "--epoch", required=True, help="epoch size", default=0)
    ap.add_argument("-m", "--model", required=True, help="model file to write(h5)", default=None)

    args = vars(ap.parse_args())
    lyr = args["layers"]
    gnd = args["ground"]
    kernel = int(args["kernelpixels"])
    sample = args["sample"]
    batch = int(args["batch"])
    epoch = int(args["epoch"])
    model = args["model"]
    source = args["dir"]
    nonzero = args["nonzero"]

    data = getData(lyr, source, kernel)
    myList = []
    for idx, ly in enumerate(lyr):
        myList = getDataSelection(ly, nonzero, 10000, gnd, 0, 0, myList)
        myList = getDataSelection(ly, nonzero, 10000, gnd, 0, 1, myList)
        myList = getDataSelection(ly, nonzero, sample[idx], gnd, 1, 0, myList)
        myList = getDataSelection(ly, nonzero, sample[idx], gnd, 1, 1, myList)
    arr = np.array(myList)
    print(arr.shape)
    arr = np.unique(arr, axis=0)
    print(arr.shape)
    data, label = getSelectedData(data, arr, gnd, kernel)
    uniquev, counts = np.unique(label, return_counts=True)
    print(np.array([uniquev, counts]).T)
    createModel(len(lyr), data, label, kernel, batch, epoch, model)
