# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import gc

import numpy as np
import argparse
import time

start = time.time()
from datetime import timedelta
from tqdm import tqdm
import tensorflow as tf
from keras.models import load_model


def predict(layers, modelfile, kernel_pixels, resultfile, index, batch_size, xsize=493, ysize=493, zsize=188):
    padding = int(kernel_pixels / 2)
    data = []
    for layer in layers:
        npa = np.load(layer)
        npa = np.pad(npa, (padding,), 'edge')
        npa = np.array(npa, dtype=np.float32)
        data.append(npa)
    ind = np.load(index)
    print("Creating Model")
    print("Number of bands:", len(layers))
    model = load_model(modelfile)
    result = np.empty((zsize, ysize, xsize))
    result[:] = np.nan
    result0 = np.empty((zsize, ysize, xsize))
    result0[:] = np.nan
    result1 = np.empty((zsize, ysize, xsize))
    result1[:] = np.nan
    lenght = int(np.ceil(len(ind) / float(batch_size)))
    strategy = tf.distribute.MirroredStrategy()
    with strategy.scope():
        for i in tqdm(range(lenght), desc="Batch Process", ascii=False, ncols=75):
            indexes = ind[i * batch_size:(i + 1) * batch_size, :]
            dt = []
            for c in range(len(indexes)):
                z = indexes[c][0]
                y = indexes[c][1]
                x = indexes[c][2]
                batch = []
                for band in data:
                    batch.append(band[z:z + kernel_pixels, y:y + kernel_pixels, x:x + kernel_pixels])
                vl = np.array(batch)
                dt.append(vl)
                del vl
            dt = np.asarray(dt, dtype=np.float32)
            pred = model.predict(dt, verbose=1)
            pred_y = pred.argmax(axis=-1)
            for j in range(len(pred_y)):
                z = indexes[j][0]
                y = indexes[j][1]
                x = indexes[j][2]
                result[z, y, x] = pred_y[j]
                result0[z, y, x] = pred[j, 0]
                result1[z, y, x] = pred[j, 1]
            gc.collect()

    np.save(resultfile + ".npy", result)
    np.save(resultfile + "0.npy", result0)
    np.save(resultfile + "1.npy", result1)


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument("-l", "--layers", required=True, help="help", nargs='+', default=[])
    ap.add_argument("-k", "--kernelpixels", required=True, help="kernel pixels size", default=0)
    ap.add_argument("-m", "--model", required=True, help="model file to read(h5)", default=None)
    ap.add_argument("-r", "--result", required=True, help="result file to write(npy)", default=None)
    ap.add_argument("-i", "--index", required=True, help="index file(npy)", default=None)
    ap.add_argument("-b", "--batch", required=True, help="batch size", default=None)
    args = vars(ap.parse_args())
    lyr = args["layers"]
    kernel = int(args["kernelpixels"])
    model = args["model"]
    result = args["result"]
    index = args["index"]
    batch = int(args["batch"])
    predict(lyr, model, kernel, result, index, batch)

end = time.time()
td = timedelta(seconds=int(end - start))
print("Total processing time", td)
